package com.secretalchemy;

import java.util.regex.*;
import java.util.ArrayList;
import java.io.*;

/** This class is a static object that will invoke GnuPG behind the scenes
 * to accomplish various user-specified actions.
 */
public class GnuPGController {
    /** A cached copy of the results of findGnuPGExecutablePath. */
    private final static String cmdPath = findGnuPGExecutablePath();
    /** A cached copy of the results of findGnuPGConfigurationPath. */
    private final static String cnfPath = findGnuPGConfigurationPath();
    /** A cached copy of the results of findGnuPGVersion. */
    private final static int version = findGnuPGVersion();

    /** Returns the path to GnuPG, or null if GnuPG could not be found.
     *
     * <ul>
     * <li>Preconditions: None</li>
     * <li>Postconditions: None</li>
     * <li>Returns: the path to GnuPG, or null if GnuPG could not be found.</li>
     * <li>Error handling: Cannot raise errors</li>
     * <li>Side effects: None</li>
     * </ul>
     *
     * @return The path to GnuPG, or null if GnuPG could not be found.
     */
    public static String getGnuPGExecutablePath() {
        return cmdPath;
    }

    /** Returns the path to gpg.conf, or 'null' if it could not be found.
     *
     * <ul>
     * <li>Preconditions: None</li>
     * <li>Postconditions: None</li>
     * <li>Returns: the path to gpg.conf, or null if it could not be found.</li>
     * <li>Error handling: Cannot raise errors</li>
     * <li>Side effects: None</li>
     * </ul>
     *
     * @return The path to gpg.conf, or 'null' if it could not be found.
     */
    public static String getGnuPGConfigurationPath() {
        return cnfPath;
    }

    /** Returns the version number of GnuPG in the form of a hexadecimal
     * value.  E.g., version 2.0.10 is hex 0x02000A.  This makes it possible
     * to test GnuPG versions with simple arithmetic comparisons: for
     * instance, 0x02000A is greater than 0x010404, so version 2.0.10 is
     * more recent than 1.4.4.
     *
     * <ul>
     * <li>Preconditions: None</li>
     * <li>Postconditions: None</li>
     * <li>Returns: The GnuPG version as a hex value</li>
     * <li>Error handling: None</li>
     * <li>Side effects: None</li>
     * </ul>
     *
     * @return The GnuPG version as a hex value
     */
    public static int getGnuPGVersion() {
        return version;
    }

    /** Discovers the version number of GnuPG in the form of a hexadecimal
     * value.  E.g., version 2.0.10 is hex 0x02000A.  This makes it possible
     * to test GnuPG versions with simple arithmetic comparisons: for
     * instance, 0x02000A is greater than 0x010404, so version 2.0.10 is
     * more recent than 1.4.4.
     *
     * <ul>
     * <li>Preconditions: None</li>
     * <li>Postconditions: None</li>
     * <li>Returns: The GnuPG version as a hex value</li>
     * <li>Error handling: Returns 0 on failure</li>
     * <li>Side effects: Accesses the disk and starts a subprocess</li>
     * </ul>
     *
     * @return The GnuPG version as a hex value
     */
    private static int findGnuPGVersion() {
        if (null == getGnuPGExecutablePath()) {
            return 0;
        }
        String line = "";
        try {
            Process p = new ProcessBuilder(getGnuPGExecutablePath(), "--version").start();
            InputStreamReader isr = new InputStreamReader(p.getInputStream());
            BufferedReader br = new BufferedReader(isr);
            line = br.readLine();
            p.destroy();
        } catch (IOException ioe) {
            return 0;
        }

        Pattern versionPtn = Pattern.compile("\\d+(\\.\\d+)+");
        Matcher m = versionPtn.matcher(line);
        m.find();
        String ver = m.group();

        /* Official GnuPG releases have a major, minor and revision
         * number.  If our version number doesn't have three components,
         * then we're not dealing with an official release and we bail
         * out. */
        if (ver == null || ver.split("\\.").length != 3) {
            return 0;
        }
        String[] versionNums = ver.split("\\.");
        int major = Integer.parseInt(versionNums[0]);
        int minor = Integer.parseInt(versionNums[1]);
        int rev   = Integer.parseInt(versionNums[2]);

        int rv = major << 8;
        rv += minor;
        rv = rv << 8;
        rv += rev;
        return rv;
    }

    /** Finds the path to gpg.conf, or null if it could not be found.  Note
     * that this process involves a lot of platform-specific hackery.
     *
     * <ul>
     * <li>Preconditions: None</li>
     * <li>Postconditions: None</li>
     * <li>Returns: the path to gpg.conf, or null if it could not be found.</li>
     * <li>Error handling: Returns null on any error</li>
     * <li>Side effects: Accesses the disk.  If the necessary directory does
     * not exist, will attempt to create it.</li>
     * </ul>
     *
     * @return The path to gpg.conf, or null if it could not be found.
     */
    private static String findGnuPGConfigurationPath() {
        String path = "";
        if (Pattern.matches("Windows.*", System.getProperty("os.name"))) {
            if (System.getenv().containsKey("GNUPGHOME")) {
                path = System.getenv("GNUPGHOME") + "\\";
            } else {
                path = System.getenv("APPDATA") + "\\gnupg\\";
            }
        }
        else {
            if (System.getenv().containsKey("GNUPGHOME")) {
                path = System.getenv("GNUPGHOME") + "/";
            } else {
                path = System.getProperty("user.home") + "/.gnupg/";
            }
        }

        File f = new File(path);
        if (! f.exists()) {
            if (! f.mkdirs()) {
                return null;
            }
        }
        return path;
    }

    /** Finds the path to GnuPG, or null if GnuPG could not be found.  Note
     * that this task involves a lot of Windows- and UNIX-specific hacks.
     * It is not a model of good "write once, run anywhere" code.
     *
     * <ul>
     * <li>Preconditions: None</li>
     * <li>Postconditions: None</li>
     * <li>Returns: the path to GnuPG, or null if GnuPG could not be found.</li>
     * <li>Error handling: Returns null on failure</li>
     * <li>Side effects: Accesses the disk</li>
     * </ul>
     *
     * @return The path to GnuPG, or null if GnuPG could not be found.
     */
    private static String findGnuPGExecutablePath() {
        if (Pattern.matches("Windows.*", System.getProperty("os.name"))) {
            try {
                Pattern ptrn = Pattern.compile(
                    "\\s+Install Directory\\s+REG_SZ\\s+([^\\s].*)");
                String regPath = "C:\\Windows\\System32\\reg.exe";

                /* On 32-bit systems GnuPG installs to HKLM\Software.
                 * On 64-bit, it installs to HKLM\Software\Wow6432Node.
                 * We could either ask the JVM which architecture we're
                 * on, or we could just ask for both keys.  Six of one,
                 * half-dozen of the other.
                 */
                for (String regkey: new String[] {
                    "HKLM\\Software\\GNU\\GnuPG",
                    "HKLM\\Software\\Wow6432Node\\GNU\\GnuPG"
                }) {
                    Process p = new ProcessBuilder(regPath, "QUERY",
                        regkey, "/v",
                        "Install Directory").start();
                    InputStream is = p.getInputStream();
                    InputStreamReader isr = new InputStreamReader(is);
                    BufferedReader br = new BufferedReader(isr);
                    String line = br.readLine();

                    while (line != null) {
                        Matcher m = ptrn.matcher(line);
                        if (m.find()) {
                            String path = m.group(1) + "\\";
                            String[] fnames = { "gpg2.exe", "gpg.exe" };
                            for (String s: fnames) {
                                if (new File(path + s).exists()) {
                                    return path + s;
                                }
                            }
                        }
                        line = br.readLine();
                    }
                }
            } catch (IOException ioe) {
            }
            return null;
        } else {
            ArrayList<String> paths = new ArrayList<String>();
            for (String s: System.getenv("PATH").split(":")) {
                paths.add(s);
            }
            paths.add("/sw/bin");
            paths.add("/opt/local/bin");
            paths.add("/usr/local/bin");
            paths.add(System.getProperty("user.home") + "/bin");

            for (String path: paths) {
                for (String s: new String[] { "gpg2", "gpg" }) {
                    File f = new File(path + "/" + s);
                    if (f.exists()) {
                        return path + "/" + s;
                    }
                }
            }
            return null;
        }
    }
}
