package com.secretalchemy;

/* Bedouin is in the public domain.  This means you can do anything
 * with it.  It also means no warranties, no promises, no guarantees.
 * Use at your own risk.
 */

import javax.swing.*;
import java.util.regex.Pattern;
import java.io.*;

public class Main {
    /** A generic test to see if we're running on any version of Windows. */
    static private boolean windowsPlatform =
            Pattern.matches("Windows.*", System.getProperty("os.name"));
    /** Gets set to true if-and-when we generate an entirely new gpg.conf
     * file, which comes with SHA-256 support.
     */
    static private boolean wroteNewFile = false;

    /** Adds lines to gpg.conf which enable SHA-256 support.
     *
     * <ul>
     * <li>Preconditions: GnuPGController must have been able to find the
     * GnuPG configuration file.</li>
     * <li>Postconditions: None</li>
     * <li>Return value: true if nothing went wrong; false otherwise</li>
     * <li>Error handling: Exceptions are reported but swallowed; no
     * recovery is attempted; any exception causes a return value of false.</li>
     * <li>Side effects: Changes the gpg.conf file.</li>
     * </ul>
     *
     * @return Whether the migration was successful
     */
    public static boolean migrate() {
        // The most basic check of our precondition: see if we can get the
        // path.  If not, then we bomb out with a false.
        String cnfPath = GnuPGController.getGnuPGConfigurationPath();
        if (null == cnfPath) {
            return false;
        }
        String cnfFn = cnfPath + "gpg.conf";
        String backFn = cnfPath + "gpg.conf.bedouin";
        while (new File(backFn).exists()) {
            backFn += ".again";
        }
        try {
            FileInputStream fis = new FileInputStream(cnfFn);
            InputStreamReader isr = new InputStreamReader(fis);
            BufferedReader br = new BufferedReader(isr);
            FileOutputStream fos = new FileOutputStream(backFn);
            OutputStreamWriter osw = new OutputStreamWriter(fos);
            PrintWriter pw = new PrintWriter(osw);
            boolean alreadyConfigured = false;

            String line = br.readLine();
            while (null != line) {
                if (Pattern.matches("# begin Bedouin autoconfiguration",
                        line)) {
                    alreadyConfigured = true;
                }
                pw.println(line);
                line = br.readLine();
            }

            pw.flush();
            fis.close();
            fos.close();

            if (alreadyConfigured) {
                return true;
            }

            fos = new FileOutputStream(cnfFn, true);
            osw = new OutputStreamWriter(fos);
            pw = new PrintWriter(osw);

            String resName = "/Resources/BedouinConf.txt";
            InputStream is = Main.class.getResourceAsStream(resName);
            isr = new InputStreamReader(is);
            br = new BufferedReader(isr);
            pw.println();
            pw.println();

            line = br.readLine();
            while (null != line) {
                pw.println(line);
                line = br.readLine();
            }
            pw.flush();
            is.close();
            fos.close();

            return true;
        } catch (Exception e) {
            JOptionPane.showMessageDialog(null,
                    "Unexpected disk error\n\n" +
                    "An unexpected disk error has occurred.  Please check\n"+
                    "to make sure the GnuPG configuration directory exists\n"+
                    "and is writable.",
                    "Bedouin Migration Assistant",
                    JOptionPane.ERROR_MESSAGE);
            return false;
        }
    }

    /** Create a simple default gpg.conf file.
     *
     * <ul>
     * <li>Precondition: GnuPGController must have been able to find the
     * configuration file directory.</li>
     * <li>Postcondition: None
     * <li>Returns: true if it was able to successfully write out the file;
     * false otherwise.</li>
     * <li>Error handling: Swallows exceptions, reports the error, and
     * returns false.</li>
     * <li>Side effects: creates a new gpg.conf file.</li>
     * </ul>
     *
     * @return True if writing the file was successful; false otherwise.
     */
    private static boolean populateConfigurationFile() {
        String resource = "/Resources/DefaultGPGConf.txt";
        InputStream is = Main.class.getResourceAsStream(resource);
        InputStreamReader isr = new InputStreamReader(is);

        String outfilename = GnuPGController.getGnuPGConfigurationPath();
        if (null == outfilename) {
            return false;
        }
        outfilename += "gpg.conf";
        try {
            FileOutputStream fos = new FileOutputStream(outfilename);
            OutputStreamWriter osw = new OutputStreamWriter(fos);
            PrintWriter pw = new PrintWriter(osw);

            BufferedReader br = new BufferedReader(isr);
            String line = br.readLine();
            while (null != line) {
                pw.println(line);
                line = br.readLine();
            }
            is.close();
            pw.flush();
            fos.close();
            wroteNewFile = true;
            return true;
        } catch (IOException ioe) {
            JOptionPane.showMessageDialog(null,
                    "Unknown disk error\n\n" +
                    "GnuPG's configuration file could not be written\n" +
                    "to disk.  Bedouin tried to write to the file:\n\n" +
                    outfilename + "\n\n" +
                    "... and was unsuccessful.",
                    "Bedouin Migration Assistant",
                    JOptionPane.ERROR_MESSAGE);
            return false;
        }
    }

    /** Determines whether a gpg.conf file already exists.  Note that it
     * does not attempt to do any checking to see if the gpg.conf file
     * has already been migrated to SHA-256.
     *
     * <ul>
     * <li>Precondition: GnuPGController can find the configuration path.</li>
     * <li>Postcondition: None</li>
     * <li>Returns: True if a gpg.conf file already exists; false otherwise</li>
     * <li>Error handling: Reports the error, swallows it, and returns false</li>
     * <li>Side effects: Accesses, but does not modify, the disk</li>
     * </ul>
     *
     * @return true if a configuration file exists; false otherwise.
     */
    private static boolean hasConfigurationFile() {
        String path = GnuPGController.getGnuPGConfigurationPath();
        if (null == path) {
            String foo = "";
            if (windowsPlatform) {
                foo = System.getenv("APPDATA") + "\\gnupg\\";
            } else {
                foo = System.getenv("HOME") + "/.gnupg/";
            }
            JOptionPane.showMessageDialog(null,
                    "Unknown disk error\n\n" +
                    "GnuPG stores important data in the folder\n" +
                    foo + "\n\n" +
                    "This folder does not exist and could not\n" +
                    "be created.",
                    "Bedouin Migration Assistant",
                    JOptionPane.ERROR_MESSAGE);
            return false;
        }
        if (! (new File(path + "gpg.conf").exists())) {
            int choice = JOptionPane.showConfirmDialog(null,
                    "No configuration file found\n\n" +
                    "The GnuPG configuration file could not be found.\n" +
                    "Migration cannot continue without this file.\n\n" +
                    "Click 'OK' to create a standard configuration file,\n" +
                    "or 'Cancel' to quit Bedouin.",
                    "Bedouin Migration Assistant",
                    JOptionPane.OK_CANCEL_OPTION,
                    JOptionPane.INFORMATION_MESSAGE);
            if (choice == JOptionPane.CANCEL_OPTION) {
                return false;
            } else {
                boolean rv = populateConfigurationFile();
                if (rv) {
                    JOptionPane.showMessageDialog(null,
                    "Configuration file created!\n\n" +
                    "A new configuration file has been generated.  You\n" +
                    "may find it at:\n\n" +
                    GnuPGController.getGnuPGConfigurationPath() +
                    "gpg.conf\n\n" +
                    "GnuPG is now set up to use SHA-256 whenever possible.",
                    "Bedouin Migration Assistant",
                    JOptionPane.INFORMATION_MESSAGE);
                }
                return rv;
                }
            }
        return true;
    }

    /** Tests to determine whether GnuPG is present on this system.
     *
     * <ul>
     * <li>Preconditions: None</li>
     * <li>Postconditions: None</li>
     * <li>Returns: true if GnuPG exists, false otherwise</li>
     * <li>Error handling: Reports and swallows the error, then returns false</li>
     * <li>Side effects: May construct a URL; may start a web browser.</li>
     * </ul>
     *
     * @return true if GnuPG exists; false otherwise.
     */
    private static boolean hasGnuPG() {
        if (null == GnuPGController.getGnuPGExecutablePath()) {
            if (windowsPlatform) {
                try {
                    String foo = "http://www.gpg4win.org/download.html";
                    java.net.URI uri = new java.net.URI(foo);

                    if (JOptionPane.YES_OPTION ==
                            JOptionPane.showConfirmDialog(null,
                            "Couldn't find GnuPG!\n\n" +
                            "You do not appear to have GnuPG installed.\n" +
                            "Click 'Yes' to visit the official site of\n" +
                            "GnuPG for Windows.  Download and install\n" +
                            "GnuPG for Windows, and then try again.",
                            "Bedouin Migration Assistant",
                            JOptionPane.YES_NO_OPTION,
                            JOptionPane.INFORMATION_MESSAGE)) {
                        java.awt.Desktop.getDesktop().browse(uri);
                    }
                } catch (Exception e) {
                    JOptionPane.showMessageDialog(null,
                            "Couldn't find GnuPG!\n\n" +
                            "GnuPG does not seem to be installed.\n\n" +
                            "Please visit http://www.gpg4win.org, download\n"+
                            "and install the latest version of gpg4win, and\n"+
                            "run Bedouin again.",
                            "Bedouin Migration Assistant",
                            JOptionPane.INFORMATION_MESSAGE);
                }
            } else { // UNIX platform
                JOptionPane.showMessageDialog(null,
                        "Couldn't find GnuPG!\n\n" +
                        "GnuPG does not seem to be installed.\n\n" +
                        "Please install GnuPG 1.4.4 or later, and run\n" +
                        "Bedouin again.",
                        "Bedouin Migration Assistant",
                        JOptionPane.INFORMATION_MESSAGE);
            }
            return false;
        }
        return true;
    }

    /** Tests to see whether GnuPG 1.4.4 or later is being used.  If
     * an old version is being used, prompts the user to upgrade.
     *
     * <ul>
     * <li>Precondition: None</li>
     * <li>Postcondition: GnuPG will be known to be at least v1.4.4.</li>
     * <li>Returns: true if GnuPG is at least v1.4.4; false otherwise</li>
     * <li>Error handling: reports and swallows error, then returns false</li>
     * <li>Side effects: invokes a subprocess</li>
     * </ul>
     *
     * @return true if GnuPG is at least v1.4.4; false otherwise
     */
    private static boolean hasRecentVersion() {
        if (GnuPGController.getGnuPGVersion() < 0x010404) {
            if (windowsPlatform) {
                try {
                    String gpg4win = "http://www.gpg4win.org/download.html";
                    java.net.URI uri = new java.net.URI(gpg4win);
                    int choice = JOptionPane.showConfirmDialog(null,
                            "Obsolete GnuPG detected\n\n" +
                            "This GnuPG version is very old and needs to be\n"+
                            "upgraded.  Click 'Yes' to open a web browser to\n"+
                            "the gpg4win website, where the latest version\n"+
                            "of GnuPG for Windows may be downloaded.",
                            "Bedouin Migration Assistant",
                            JOptionPane.YES_NO_OPTION,
                            JOptionPane.INFORMATION_MESSAGE);
                    if (choice == JOptionPane.YES_OPTION) {
                        java.awt.Desktop.getDesktop().browse(uri);
                    }
                } catch (Exception ioe) {
                    JOptionPane.showMessageDialog(null,
                            "Obsolete GnuPG detected\n\n" +
                            "This GnuPG version is very old and needs to be\n"+
                            "upgraded.  Please visit http://www.gpg4win.org\n"+
                            "and install the most recent version of GnuPG\n"+
                            "for Windows.",
                            "Bedouin Migration Assistant",
                            JOptionPane.INFORMATION_MESSAGE);
                }
            } else {
                JOptionPane.showMessageDialog(null,
                        "Obsolete GnuPG detected\n\n" +
                        "This GnuPG version is very old and needs to be\n" +
                        "upgraded.  Please contact your vendor and upgrade\n" +
                        "to GnuPG 1.4.4 or newer.",
                        "Bedouin Migration Assistant",
                        JOptionPane.INFORMATION_MESSAGE);
            }
            return false;
        }
        return true;
    }

    public static void main(String[] args) {
        try {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        } catch (Exception e) {
            // do nothing: we just use Metal
        }

        if (! (hasGnuPG() && hasRecentVersion() && hasConfigurationFile()))
            return;
        if (wroteNewFile)
            return;

        Migrator m = new Migrator();
        m.setVisible(true);
        
    }
}